# frozen_string_literal: true

require_relative "miradore/version"
require "httparty"
require "json"
require "crack/xml"
require "finest/builder"

module Ruby
  module Miradore

    def self.url
      {
        item_v1: "https://%<subdomain>s.online.miradore.com/API/%<item>s?auth=%<auth>s&select=*,%<attribute>s&filters=%<filter>s",
        request_v1: "https://%<subdomain>s.online.miradore.com/API/Device?auth=%<auth>s&select=*,Category.*,InvDevice.*,InvStorage.*,InvOS.*,Enrollment.*,Client.*,User.*,Location.*,Security.*,ReportedLocation.*&filters=%<filter>s",
        v2: "https://%<subdomain>s.online.miradore.com/API/v2/Device/%<id>s/%<method>s"
      }
    end

    class Request
      include Finest::Builder
      include HTTParty
      format :json

      attr_writer :subdomain, :auth

      def initialize(json = {})
        super json
        @subdomain ||= json[:subdomain]
        @auth ||= json[:auth]
        remove_instance_variable(:@subdomain) if @subdomain.nil?
        remove_instance_variable(:@auth) if @auth.nil?
      end

      #
      # This call will perform all GET calls regardless the API version that it's calling.
      # The version would be determined based on the superclass name.
      # If the superclass is Request the GET will call simple_v1 which will require the item which is class name.
      # otherwise will call to retrieve devices based on filters.
      #
      def self.all(**args)
        response = JSON.parse(
          Crack::XML.parse(
            get(
              Miradore.url[version] % args.merge(
                item: object_name,
                attribute: args.fetch(:attribute, "*"),
                auth: args.fetch(:auth, nil),
                filter: args.fetch(:filter, nil)
              )
            ).body
          ).to_json,
          quirks_mode: true
        ).dig("Content", "Items")
        transform(response, args)
      rescue StandardError => e
        MiradoreError.new({ error: e&.message || "Error connecting to manager service provider", status: :not_found })
      end

      class << self
        private

        def transform(res, args = {})
          instance_name = "Ruby::Miradore::#{res.keys.first.capitalize}"
          if res.values.first.is_a?(Array)
            return res.values.first.map! do |e|
              Object.const_get(instance_name).new(args.merge(e.transform_keys(&:downcase)))
            end
          end

          Object.const_get(instance_name).new(args.merge(res.values.first&.transform_keys(&:downcase)))
        rescue StandardError
          MiradoreError.new({ error: "#{object_name}(s) not found", status: :not_found })
        end

        def version
          superclass == Object ? :request_v1 : :item_v1
        end

        def object_name
          to_s.split("::").last
        end
      end

    end

    class Device < Request
      include HTTParty
      format :json
      headers "Content-Type": "application/json"
      headers "Accept": "application/json"

      def lock(**args)
        http_method(args.merge(method: :post, id: id, action: __method__))
      end

      def lostmode(**args)
        if args[0]
          http_method(args.merge(method: :post, id: id, action: __method__))
        else
          http_method(args.merge(method: :delete, id: id, action: __method__))
        end
      end

      def reboot(**args)
        http_method(args.merge(method: :post, id: id, action: __method__))
      end

      def wipe(**args)
        http_method(args.merge(method: :post, id: id, action: __method__))
      end

      def location(**args)
        http_method(args.merge(method: :get, action: __method__))
      end

      def patch(**args)
        http_method(args.merge(method: :patch))
      end

      private

      def http_method(**args)
        self.class.headers "X-API-Key" => @auth || args[:auth]
        url = Miradore.url[:v2] % {
          subdomain: @subdomain || args[:subdomain],
          id: args.fetch(:id, id),
          method: args.fetch(:action, nil)
        }
        self.class.method(args.fetch(:method, nil)).call(url, body: JSON.generate(args.fetch(:body, {})))
      rescue StandardError => e
        { error: e&.message || "Error connecting to manager service provider", status: :not_found }
      end

    end

    class User < Request; end

    class Organization < Request; end

    class MiradoreError < Request; end
  end
end
